unit main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, ComCtrls, Jpeg, Math;

type
  PRGB24 = ^TRGB24;
  TRGB24 = record B, G, R: Byte; end;
  PRGBArray = ^TRGBArray;
  TRGBArray = array [Word] of TRGB24;

  TfrmCurves = class(TForm)
    pnlTool: TPanel;
    imgCurve: TImage;
    imgView: TImage;
    StatusBar1: TStatusBar;
    dlgOpenImg: TOpenDialog;
    btnOpen: TButton;
    btnReset: TButton;
    Bevel1: TBevel;
    cbPreset: TComboBox;
    labPreset: TLabel;
    btnAdd: TButton;
    cbChannel: TComboBox;
    labChannel: TLabel;
    lblIn: TLabel;
    lblOut: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure imgCurveMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure imgCurveMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure imgCurveMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure btnOpenClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure rbRedClick(Sender: TObject);
    procedure btnResetClick(Sender: TObject);
    procedure imgViewMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure imgViewMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure cbPresetChange(Sender: TObject);
    procedure cbChannelChange(Sender: TObject);
    procedure btnAddClick(Sender: TObject);
    procedure chkAltClick(Sender: TObject);
  private
    procedure SetPandU;
    function GetCurvePoint(i: Integer; v: Single): Single;
    procedure DrawCurve;
    procedure AddPoint(pt: TPoint);
    procedure DelPoint(idx: Integer);
    procedure MakePreviewImage;
    procedure ApplyCurve(Img: TBitmap);
    procedure GetHist;
    procedure GetCurvesPreset;
    procedure LoadPreset(Idx: Integer);
    { Private declarations }
  public
    { Public declarations }
  end;

var
  frmCurves: TfrmCurves;

implementation

{$R *.dfm}

var
  nPts: Array[0..3] of Integer;
  ptX, ptY: Array[0..3, 1..32] of Integer;
  ptP, ptU: Array[0..3, 1..32] of Single;
  nHist: Array[0..3, Byte] of Integer;
  maxHist: Array[0..3] of Integer;
  LUT: Array[0..3, 0..255] of Byte;
  aPt, ChnIdx: Integer;
  TheImage, Img: TBitmap;
  ImgLoaded: Boolean;
  CurvesPath: String;

function CompareNatural(s1, s2: string): Integer;
  function ExtractNr(n: Integer; var Txt: string): Int64;
  begin
    while (n <= Length(Txt)) and (Txt[n] >= '0') and (Txt[n] <= '9') do n := n + 1;
    Result := StrToInt64Def(Copy(Txt, 1, n - 1), 0);
    Delete(Txt, 1, (n - 1));
  end;
var
  b: Boolean;
begin
  Result := 0;
  s1 := LowerCase(s1);
  s2 := LowerCase(s2);
  if (s1 <> s2) and (s1 <> '') and (s2 <> '') then begin
    b := False;
    while (not b) do begin
      if ((s1[1] >= '0') and (s1[1] <= '9')) and ((s2[1] >= '0') and (s2[1] <= '9')) then
        Result := Sign(ExtractNr(1, s1) - ExtractNr(1, s2))
      else
        Result := Sign(Integer(s1[1]) - Integer(s2[1]));
      b := (Result <> 0) or (Min(Length(s1), Length(s2)) < 2);
      if not b then begin
        Delete(s1, 1, 1);
        Delete(s2, 1, 1);
      end;
    end;
  end;
  if Result = 0 then begin
    if (Length(s1) = 1) and (Length(s2) = 1) then
      Result := Sign(Integer(s1[1]) - Integer(s2[1]))
    else
      Result := Sign(Length(s1) - Length(s2));
  end;
end;

function SortMe(List: TStringList; i1, i2: Integer): Integer;
begin
  Result := CompareNatural(List[i1], List[i2]);
end;

function CalcImgSize(w, h, tw, th: integer): TPoint;
begin
  Result:= Point(0, 0);
  if (w = 0) or (h = 0) then Exit;
  if (w < tw) and (h < th) then Result:= Point(w, h)
  else begin
    if w > h then begin
      if w < tw then tw := w;
      Result:= Point(tw, Trunc(tw * h / w));
      if Result.Y > th then Result:= Point(Trunc(th * w / h), th);
    end else begin
      if h < th then th := h;
      Result:= Point(Trunc(th * w / h), th);
      if Result.X > tw then Result:= Point(tw, Trunc(tw * h / w));
    end;
  end;
end;

function Blend(Color1, Color2: TColor; A: Byte): TColor;
var
  c1, c2: LongInt;
  r, g, b, v1, v2: byte;
begin
  A := Round(2.56 * A);
  c1 := ColorToRGB(Color1);
  c2 := ColorToRGB(Color2);
  r := A * (Byte(c1) - Byte(c2)) shr 8 + Byte(c2);
  g := A * (Byte(c1 shr 8) - Byte(c2 shr 8)) shr 8 + Byte(c2 shr 8);
  b := A * (Byte(c1 shr 16) - Byte(c2 shr 16)) shr 8 + Byte(c2 shr 16);
  Result := (b shl 16) + (g shl 8) + r;
end;

procedure BilinearRescale(Src, Dest: TBitmap);
var
  x, y, px, py: Integer;
  i, x1, x2, z, z2, iz2: Integer;
  w1, w2, w3, w4: Integer;
  Ratio: Integer;
  sDst, sDstOff: Integer;
  PScanLine: array of PRGBArray;
  Src1, Src2: PRGBArray;
  C, C1, C2: TRGB24;
begin
  if (Dest.Width < 2) or (Dest.Height < 2) then begin
    Dest.Assign(Src);
    Exit;
  end;
  SetLength(PScanLine, Src.Height);
  PScanLine[0]:= (Src.Scanline[0]);
  i := Integer(Src.Scanline[1]) - Integer(PScanLine[0]);
  for y := 1 to Src.Height - 1 do PScanLine[y]:= PRGBArray(Integer(PScanLine[y - 1]) + i);
  sDst := Integer(Dest.ScanLine[0]);
  sDstOff := Integer(Dest.ScanLine[1]) - sDst;
  Ratio := ((Src.Width - 1) shl 15) div Dest.Width;
  py := 0;
  for y := 0 to Dest.Height - 1 do begin
    i := py shr 15;
    if i > src.Height - 1 then i := src.Height - 1;
    Src1 := PScanline[i];
    if i < src.Height - 1 then Src2 := PScanline[i + 1] else Src2 := Src1;
    z2 := py and $7FFF;
    iz2 := $8000 - z2;
    px := 0;
    for x := 0 to Dest.Width - 1 do begin
      x1 := px shr 15;
      x2 := x1 + 1;
      C1 := Src1[x1];
      C2 := Src2[x1];
      z := px and $7FFF;
      w2 := (z * iz2) shr 15;
      w1 := iz2 - w2;
      w4 := (z * z2) shr 15;
      w3 := z2 - w4;
      C.R := (C1.R * w1 + Src1[x2].R * w2 + C2.R * w3 + Src2[x2].R * w4) shr 15;
      C.G := (C1.G * w1 + Src1[x2].G * w2 + C2.G * w3 + Src2[x2].G * w4) shr 15;
      C.B := (C1.B * w1 + Src2[x2].B * w2 + C2.B * w3 + Src2[x2].B * w4) shr 15;
      PRGBArray(sDst)[x] := C;
      Inc(px, Ratio);
    end;
    sDst := sDst + SDstOff;
    Inc(py, Ratio);
  end;
  SetLength(PScanline, 0);
end;

function PtInCircle(cx, cy, x, y, radius: Integer): Boolean;
begin
  Result:= ((cx - x) * (cx - x)) + ((cy - y) * (cy - y)) <= radius * radius;
end;

procedure WuLine(Src: TBitmap; x1, y1, x2, y2: integer; Color: TColor);
var
  c: Cardinal;
  r, g, b: Byte;
  i, dx, dy, x, y, w, h, a1, a2 : integer;
  dxi, dyi, gradient : integer;
  Line: array of PRGBArray;

  function BlendPixel(x, y, a: Integer): TRGB24;
  begin
    Result.R:= a * (r - Line[y][x].R) shr 8 + Line[y][x].R;
    Result.G:= a * (g - Line[y][x].G) shr 8 + Line[y][x].G;
    Result.B:= a * (b - Line[y][x].B) shr 8 + Line[y][x].B;
  end;

begin
  c:= ColorToRGB(Color);
  r:= c and 255;
  g:= (c shr 8) and 255;
  b:= (c shr 16) and 255;
  w:= Src.Width;
  h:= Src.Height;
  if (x1 = x2) or (y1 = y2) then begin
    Src.Canvas.Pen.Color:= Color;
    Src.Canvas.MoveTo(x1, y1);
    Src.Canvas.LineTo(x2, y2);
    Exit;
  end;
  // make an array of source scanlines to speed up the rendering
  SetLength(Line, Src.Height);
  Line[0]:= (Src.Scanline[0]);
  i:= Integer(Src.Scanline[1]) - Integer(Line[0]);
  for y:= 1 to Src.Height - 1 do Line[y]:= PRGBArray(Integer(Line[y - 1]) + i);
  dx:= abs(x2 - x1);
  dy:= abs(y2 - y1);
  if dx > dy then begin // horizontal or vertical
    if y2 > y1 then dy:= -dy;
    gradient:= dy shl 8 div dx;
    if x2 < x1 then begin
      i:= x1; x1:= x2; x2:= i;
      dyi:= y2 shl 8;
    end else begin
      dyi:= y1 shl 8;
      gradient:= -gradient;
    end;
    if x1 >= W then x2:= W - 1;
    for x := x1 to x2 do begin
      Y:= dyi shr 8;
      if (x < 0) or (y < 0) or (y > h - 2) then Inc(dyi, gradient) else begin
        a1 := dyi - y shl 8;
        a2 := 256 - a1;
        Line[y][x]:= BlendPixel(x, y, a1);
        Line[y + 1][x]:= BlendPixel(x, y + 1, a2);
        Inc(dyi, gradient);
      end;
    end;
  end else begin
    if x2 > x1 then dx:= -dx;
    gradient:= dx shl 8 div dy;
    if y2 < y1 then begin
      i:= y1; y1:= y2; y2:= i;
      dxi:= x2 shl 8;
    end else begin
      dxi:= x1 shl 8;
      gradient:= -gradient;
    end;
    if y2 >= h then y2:= h - 1;
    for y := y1 to y2 do begin
      x:= dxi shr 8;
      if (y < 0) or (x < 0) or (x > w - 2) then Inc(dxi, gradient) else begin
        a1 := dxi - x shl 8;
        a2 := 256 - a1;
        Line[y][x]:= BlendPixel(x, y, a2);
        Line[y][x + 1]:= BlendPixel(x + 1, y, a1);
        Inc(dxi, gradient);
      end;
    end;
  end;
end;

procedure TfrmCurves.SetPandU;
var
  i: Integer;
  d, w: array[1..32] of Single;
begin
  for i:= 2 to nPts[ChnIdx] - 1 do d[i]:= 2 * (ptX[ChnIdx, i + 1] - ptX[ChnIdx, i - 1]);
  for i:= 1 to nPts[ChnIdx] - 1 do ptU[ChnIdx, i]:= ptX[ChnIdx, i + 1] - ptX[ChnIdx, i];
  for i:= 2 to nPts[ChnIdx] - 1 do w[i]:= 6 * ((ptY[ChnIdx, i + 1] - ptY[ChnIdx, i]) /
    ptU[ChnIdx, i] - (ptY[ChnIdx, i] - ptY[ChnIdx, i - 1]) / ptU[ChnIdx, i - 1]);
  for i:= 2 to nPts[ChnIdx] - 2 do begin
    w[i + 1]:= w[i + 1] - w[i] * ptU[ChnIdx, i] / d[i];
    d[i + 1]:= d[i + 1] - ptU[ChnIdx, i] * ptU[ChnIdx, i] / d[i];
  end;
  ptP[ChnIdx, 1]:= 0;
  for i:= nPts[ChnIdx] - 1 downto 2 do ptP[ChnIdx, i]:= (w[i] - ptU[ChnIdx, i] * ptP[ChnIdx, i + 1]) / d[i];
  ptP[ChnIdx, nPts[ChnIdx]]:= 0;
end;

function TfrmCurves.GetCurvePoint(i: Integer; v: Single) : Single;
var
  t0, t1: Single;
begin
  t0:= (v - ptX[ChnIdx, i]) / ptU[ChnIdx,i];
  t1:= 1 - t0;
  Result:= t0 * ptY[ChnIdx, i + 1] + t1 * ptY[ChnIdx, i] + ptU[ChnIdx, i] *
    ptU[ChnIdx, i] * ((t0*t0*t0-t0) * ptP[ChnIdx, i + 1] + (t1*t1*t1-t1) * ptP[ChnIdx, i]) / 6;
end;

procedure TfrmCurves.AddPoint(pt: TPoint);
var
  i, x: Integer;
begin
  i:= 1;
  while (i <= nPts[ChnIdx]) and (pt.X > ptX[ChnIdx, i]) do i:= i + 1;
  if i <= nPts[ChnIdx] + 1 then begin
    for x:= 31 downto i do begin
      ptX[ChnIdx, x + 1]:= ptX[ChnIdx, x];
      ptY[ChnIdx, x + 1]:= ptY[ChnIdx, x];
    end;
    ptX[ChnIdx, i]:= pt.X;
    ptY[ChnIdx, i]:= pt.Y;
    apt:= i;
    nPts[ChnIdx]:= nPts[ChnIdx] + 1;
  end;
end;

procedure TfrmCurves.DelPoint(idx: Integer);
var
  x: Integer;
begin
  if nPts[ChnIdx] = 2 then begin
    ShowMessage('At least two points must exist.');
    Exit;
  end;
  if (idx > 0) and (idx <= nPts[ChnIdx]) then begin
    for x:= idx to 31 do begin
      ptX[ChnIdx, x]:= ptX[ChnIdx, x + 1];
      ptY[ChnIdx, x]:= ptY[ChnIdx, x + 1];
    end;
    apt:= -1;
    nPts[ChnIdx]:= nPts[ChnIdx] - 1;
    DrawCurve;
  end;
end;

function ValidPoint(x, y: Integer): Boolean;
begin
  Result:= ((x > 7) and (x < 264)) and ((y > 7) and (y < 264));
end;

procedure TfrmCurves.imgCurveMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  i: Integer;
  p: TPoint;
begin
  if not ValidPoint(x, y) then Exit;

  p.X:= x - 8;
  p.Y:= 255 - (y - 8);

  apt:= -1;
  i:= 1;
  while (i <= nPts[ChnIdx]) and (ptX[ChnIdx, i] <> p.X) do inc(i);
  if i > nPts[ChnIdx] then begin
    i:= 1;
    while (i <= nPts[ChnIdx]) and (not PtInCircle(ptX[ChnIdx, i], ptY[ChnIdx, i], p.X, p.Y, 5)) do inc(i);
  end;
  if i <= nPts[ChnIdx] then apt:= i;
  if (ssLeft in Shift) then begin
    if (apt <> -1) then begin
      if (apt > 0) and (apt < nPts[ChnIdx]) then begin
        if (p.X > ptX[ChnIdx, apt - 1]) and (p.X < ptX[ChnIdx, apt + 1]) then ptX[ChnIdx, apt]:= p.X;
        ptY[ChnIdx, apt]:= p.Y;
      end;
    end else if nPts[ChnIdx] < 31 then AddPoint(p);
  end else if (ssRight in Shift) then DelPoint(apt);
  DrawCurve;
end;

procedure TfrmCurves.imgCurveMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  i: Integer;
  p: TPoint;
begin
  p.X:= x - 8;
  p.Y:= 255 - (y - 8);
  if p.X < 0 then p.X:= 0 else if p.X > 255 then p.X:= 255;
  if p.Y < 0 then p.Y:= 0 else if p.Y > 255 then p.Y:= 255;
  lblIn.Caption:= 'In: ' + IntToStr(p.X);
  lblOut.Caption:= 'Out: ' +IntToStr(LUT[ChnIdx, p.X]);

  if (not (ssLeft in Shift)) then Exit;
  if apt = -1 then begin
    i:= 1;
    while (i <= nPts[ChnIdx]) and (not PtInCircle(ptX[ChnIdx, i], ptY[ChnIdx, i], p.X, p.Y, 5)) do inc(i);
    if i <= nPts[ChnIdx] then apt:= i;
  end;
  if (ssLeft in Shift) and (apt <> -1) then begin
    if (apt > 1) and (apt < nPts[ChnIdx]) then begin
      if (p.X > ptX[ChnIdx, apt - 1]) and (p.X < ptX[ChnIdx, apt + 1]) then ptX[ChnIdx, apt]:= p.X;
    end else begin
      if (apt = 1) and (p.X < ptX[ChnIdx, apt + 1]) then ptX[ChnIdx, apt]:= p.X;
      if (apt = nPts[ChnIdx]) and (p.X > ptX[ChnIdx, apt - 1]) then ptX[ChnIdx, apt]:= p.X;
    end;
    ptY[ChnIdx, apt]:= p.Y;
  end;
  DrawCurve;
end;

procedure TfrmCurves.imgCurveMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  apt:= -1;
  DrawCurve;
end;

procedure TfrmCurves.imgViewMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  imgView.Picture.Bitmap.Assign(Img);
  imgview.Refresh;
end;

procedure TfrmCurves.imgViewMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  imgView.Picture.Bitmap.Assign(Img);
  ApplyCurve(imgView.Picture.Bitmap);
end;

procedure TfrmCurves.ApplyCurve(Img: TBitmap);
var
  SRow: PRGBArray;
  SFill, X, Y: Integer;
  Src: TBitmap;
begin
  if not ImgLoaded then Exit;
  Src:= TBitmap.Create;
  Src.Assign(Img);
  SRow:= PRGBArray(Src.ScanLine[0]);
  SFill := Integer(Src.ScanLine[1]) - Integer(SRow);
  for Y := 0 to Src.Height - 1 do begin
    for X := 0 to Src.Width - 1 do begin
      SRow[X].R:= LUT[0, LUT[1, SRow[X].R]];
      SRow[X].G:= LUT[0, LUT[2, SRow[X].G]];
      SRow[X].B:= LUT[0, LUT[3, SRow[X].B]];
    end;
    Inc(Integer(SRow), SFill);
  end;
  Img.Assign(Src);
  Src.Free;
end;

procedure TfrmCurves.GetHist;
var
  SRow: PRGBArray;
  x, y, SFill: Integer;
  Src: TBitmap;
  RGB: TRGB24;
  r, g, b, l: Byte;
begin
  if not ImgLoaded then Exit;
  for y:= 0 to 3 do begin
    maxHist[y]:= 0;
    for x := 0 to 255 do nHist[y, x]:= 0;
  end;
  Src:= imgView.Picture.Bitmap;
  SRow:= PRGBArray(Src.ScanLine[0]);
  SFill := Integer(Src.ScanLine[1]) - Integer(SRow);
  for Y := 0 to Src.Height - 1 do begin
    for X := 0 to Src.Width - 1 do begin
      rgb:= SRow[X];
      r:= RGB.R; g:= RGB.G;  b:= RGB.B; l:= (r + g + b) div 3;
      nHist[0, l]:= nHist[0, l] + 1;
      nHist[1, r]:= nHist[1, r] + 1;
      nHist[2, g]:= nHist[2, g] + 1;
      nHist[3, b]:= nHist[3, b] + 1;
    end;
    Inc(Integer(SRow), SFill);
  end;
  for y := 0 to 3 do for x := 0 to 255 do if nHist[y, x] > maxHist[y] then maxHist[y]:= nHist[y, x];
end;

procedure TfrmCurves.DrawCurve;
const
  Offset = 8;
  MaxVal = 255 + Offset;
var
  n, i, j, k, x, y, x1, x2, xpos, ypos: Integer;
  Canvas: TCanvas;
  c, c1: TColor;
  f: Single;
begin
  SetPandU;
  Canvas:= imgCurve.Picture.Bitmap.Canvas;
  Canvas.Brush.Color:= clWhite;
  Canvas.FillRect(Canvas.ClipRect);

  // Paint histogram
  case ChnIdx of
    0: c:= RGB(192, 192, 192);
    1: c:= RGB(255, 190, 190);
    2: c:= RGB(190, 220, 190);
    3: c:= RGB(190, 190, 255);
    else c:= RGB(190, 190, 255);
  end;

  c1:= Blend(c, clWhite, 30);
  Canvas.Pen.Color:= c;
  j:= MulDiv(nHist[ChnIdx, 0], 230, maxHist[ChnIdx]);
  Canvas.MoveTo(Offset, MaxVal);
  Canvas.LineTo(Offset, MaxVal - j);
  k:= (MaxVal) - j;
  for i := 1 to 255 do begin
    j:= (MaxVal) - MulDiv(nHist[ChnIdx, i], 230, maxHist[ChnIdx]);
    Canvas.Pen.Color:= c1;
    Canvas.MoveTo(i + Offset, MaxVal);
    Canvas.LineTo(i + Offset, j);
    WuLine(imgCurve.Picture.Bitmap, Offset + (i - 1), k, Offset + i, j, c);
    k:= j;
  end;
  // Histogram done...

  // Paint guidelines
  Canvas.Pen.Color:= clGray;
  Canvas.Brush.Style:= bsClear;
  Canvas.Rectangle(Rect(Offset, Offset, 256 + Offset, 256 + Offset));
  Canvas.Pen.Color:= clSilver;
  Canvas.Pen.Style:= psDot;
  for i:= 1 to 3 do begin
    j:= Offset + (i * 64);
    Canvas.MoveTo(8, j); Canvas.LineTo(256 + Offset, j);
    Canvas.MoveTo(j, 8); Canvas.LineTo(j, 256 + Offset);
  end;
  Canvas.Pen.Style:= psSolid;
  Canvas.Pen.Color:= clSilver;
  Canvas.MoveTo(Offset, MaxVal);
  Canvas.LineTo(MaxVal, Offset);
  // Guidelines done...

  // Paint points and curve
  Canvas.Pen.Color:= clBlack;
  Canvas.Brush.Color:= clBlack;
  Canvas.Brush.Style:= bsClear;
  for i:= 1 to nPts[ChnIdx] do
    Canvas.Rectangle(Offset + ptX[ChnIdx, i] - 3, MaxVal - ptY[ChnIdx, i] - 3,
      Offset + ptX[ChnIdx, i] + 4, MaxVal - ptY[ChnIdx, i] + 4);

  xpos:= 0; ypos:= 0;
  x:= Offset + ptX[ChnIdx, 1];
  y:= MaxVal - ptY[ChnIdx, 1];
  Canvas.MoveTo(Offset, y);
  if ptX[ChnIdx, 1] > 0 then Canvas.LineTo(x, y);
  for i:= 1 to nPts[ChnIdx] - 1 do begin
    x1:= ptX[ChnIdx, i];
    x2:= ptX[ChnIdx, i + 1];
    for j:=  x1 to x2 do begin
      xpos:= j;
      ypos:= Trunc(GetCurvePoint(i, xpos));
      if ypos < 0 then ypos:= 0 else if ypos > 255 then ypos:= 255;
      WuLine(imgCurve.Picture.Bitmap, x, y, Offset + xpos, MaxVal - ypos, clBlack);
      LUT[ChnIdx, xpos]:= ypos;
      x:= Offset + xpos;
      y:= MaxVal - ypos;
    end;
  end;
  Canvas.MoveTo(Offset + xpos, MaxVal - ypos);
  if ptX[ChnIdx, nPts[ChnIdx]] < 255 then Canvas.LineTo(MaxVal, MaxVal - ptY[ChnIdx, nPts[ChnIdx]]);
  if ptX[ChnIdx, 1] > 0 then for i:= 0 to ptx[ChnIdx, 1] - 1 do Lut[ChnIdx, i]:= ptY[ChnIdx, 1];
  if ptX[ChnIdx, nPts[ChnIdx]] < 255 then for i:= ptx[ChnIdx, nPts[ChnIdx]] + 1 to 255 do Lut[ChnIdx, i]:= ypos;
  // Curve and points done...

  imgView.Picture.Bitmap.Assign(Img);
  ApplyCurve(imgView.Picture.Bitmap);
end;

procedure TfrmCurves.MakePreviewImage;
var
  pt: TPoint;
begin
  pt := CalcImgSize(TheImage.Width, TheImage.Height, imgView.Width, imgView.Height);
  if (pt.X < 2) or (pt.Y < 2) then Exit;
  Img.Width:= pt.X;
  Img.Height:= pt.Y;
  BilinearRescale(TheImage, Img);
  imgView.Picture.Bitmap.Assign(Img);
  DrawCurve;
end;

procedure TfrmCurves.rbRedClick(Sender: TObject);
begin
  ChnIdx:= (Sender as TRadioButton).Tag;
  DrawCurve;
end;

procedure TfrmCurves.btnAddClick(Sender: TObject);
var
  i, j, k: Integer;
  txt: String;
  acv: Array[0..255] of Byte;
  Stream: TFileStream;
  ACVFile: String;
begin
  txt:= InputBox('Give the curve a name', 'Please type a name ', 'User');
  for i := 0 to 255 do acv[i]:= 0;
  if txt <> '' then begin
    j:= 1;
    acv[j]:= 4; inc(j, 2);
    acv[j]:= 5; inc(j, 2);
    for i := 0 to 3 do begin
      acv[j]:= nPts[i]; inc(j, 2);
      for k:= 1 to nPts[i] do begin
        acv[j]:= ptY[i, k]; inc(j, 2);
        acv[j]:= ptX[i, k]; inc(j, 2);
      end;
    end;
    acv[j]:= 2; inc(j, 2);
    acv[j]:= 0; inc(j, 2);
    acv[j]:= 0; inc(j, 2);
    acv[j]:= 255; inc(j, 2);
    acv[j]:= 255; inc(j, 2);
    ACVFile:= CurvesPath + txt + '.acv';
    Stream:= TFileStream.Create(ACVFile, fmCreate or fmOpenWrite or fmShareDenyWrite);
    try
      Stream.WriteBuffer(acv, j - 1);
    finally
      Stream.Free;
    end;
    if cbPreset.Items.IndexOf(txt) <> -1 then cbPreset.Items.Add(txt);
    cbPreset.ItemIndex:= cbPreset.Items.IndexOf(txt);
  end;
end;

procedure TfrmCurves.btnOpenClick(Sender: TObject);
begin
  if dlgOpenImg.Execute then begin
    imgView.Picture.LoadFromFile(dlgOpenImg.Filename);
    TheImage.Assign(imgView.Picture.Graphic);
    TheImage.PixelFormat := pf24bit;
    imgView.Picture.Bitmap.Assign(TheImage);
    MakePreviewImage;
    ImgLoaded := True;
    imgView.Enabled := ImgLoaded;
    GetHist;
    DrawCurve;
  end;
end;

procedure TfrmCurves.btnResetClick(Sender: TObject);
var
  j: Integer;
begin
  aPt:= -1;
  // Reset values
  nPts[ChnIdx]:= 2;
  ptX[ChnIdx, 1]:= 0;
  ptX[ChnIdx, 2]:= 255;
  ptY[ChnIdx, 1]:= 0;
  ptY[ChnIdx, 2]:= 255;
  for j:= 0 to 255 do LUT[ChnIdx, j]:= j;
  DrawCurve;
end;

procedure TfrmCurves.cbChannelChange(Sender: TObject);
begin
  if cbChannel.ItemIndex = 3 then ChnIdx:= 0 else ChnIdx:= cbChannel.ItemIndex + 1;
  DrawCurve;
end;

procedure TfrmCurves.cbPresetChange(Sender: TObject);
begin
  if cbPreset.ItemIndex <> -1 then LoadPreset(cbPreset.ItemIndex);
end;

procedure TfrmCurves.chkAltClick(Sender: TObject);
begin
  MakePreviewimage;
end;

procedure TfrmCurves.LoadPreset(Idx: Integer);
var
  x, y, i, j, k, m, n, p, q, cnt: Integer;
  FileArr, Simple: Array[0..255] of Byte;
  Stream: TFileStream;
  ACVFile: String;
begin
  q:= ChnIdx;
  ACVFile:= CurvesPath + cbPreset.Items[Idx] + '.acv';
  Stream:= TFileStream.Create(ACVFile, fmOpenRead or fmShareDenyWrite);
  try
    i:= Stream.Size;
    Stream.ReadBuffer(FileArr, i);
  finally
    Stream.Free;
  end;
  i:= i div 2;
  n:= 1;
  for j:= 0 to i - 1 do begin
    Simple[j]:= FileArr[n];
    n:= n + 2;
  end;
  if (Simple[0] <> 4) then begin
    ShowMessage('This file version is not supported. Sorry!');
    Exit;
  end;
  // Clear old values
  for i := 0 to 3 do begin
    for n := 1 to 32 do begin
      ptP[i, n]:= 0;
      ptU[i, n]:= 0;
    end;
  end;
  k:= 2;
  for m := 0 to 3 do begin
    cnt:= Simple[k];
    i:= Simple[k] * 2;
    nPts[m]:= cnt;
    n:= k + 1;
    for j:= 1 to cnt do begin
      ptX[m, j]:= Simple[n + 1];
      ptY[m, j]:= Simple[n];
      n:= n + 2;
    end;
    ChnIdx:= m;
    SetPandU;
    for p:= 1 to cnt - 1 do begin
      for x:=  ptX[m, p] to ptX[m, p + 1] do begin
        y:= Trunc(GetCurvePoint(p, x));
        if y < 0 then y:= 0 else if y > 255 then y:= 255;
        LUT[m, x]:= y;
      end;
    end;
    if ptX[m, 1] > 0 then for p:= 0 to ptx[m,  1] - 1 do Lut[m, p]:= ptY[m, 1];
    if ptX[m, nPts[m]] < 255 then for p:= ptx[m, nPts[m]] + 1 to 255 do LUT[m, p]:= 255;
    k:= k + i + 1;
  end;
  ChnIdx:= q;
  DrawCurve;
end;

procedure TfrmCurves.GetCurvesPreset;
var
  Files: TStringList;
  SR: TSearchRec;
  i: Integer;
begin
  // If directory exists then load presets else just create the directory
  if not DirectoryExists(CurvesPath) then CreateDir(CurvesPath) else begin
    cbPreset.Items.Clear;
    Files:= TStringList.Create;
    try
      if FindFirst(CurvesPath + '*.*', faArchive, SR) = 0 then begin
        repeat
          Files.Add(SR.Name);
        until FindNext(SR) <> 0;
        FindClose(SR);
      end;
      if Files.Count > 0 then begin
        Files.CustomSort(SortMe);
        cbPreset.Items.Clear;
        for i := 0 to Files.Count - 1 do cbPreset.Items.Add(ChangeFileExt(Files[i], ''));
      end;
    finally
      Files.Free;
    end;
  end;
end;

procedure TfrmCurves.FormCreate(Sender: TObject);
var
  bmp: TBitmap;
  i, j: Integer;
begin
  TheImage := TBitmap.Create;
  Img := TBitmap.Create;
  Img.PixelFormat := pf24bit;
  imgView.Picture.Bitmap.Width := 1;
  imgView.Picture.Bitmap.Height := 1;
  imgView.Picture.Bitmap.Canvas.Pixels[0, 0] := $00C4C4C4;
  bmp:= TBitmap.Create;
  bmp.PixelFormat:= pf24bit;
  bmp.Width:= 272;
  bmp.Height:= 272;
  imgCurve.Picture.Bitmap.Assign(bmp);
  bmp.Free;
  ChnIdx:= 0;
  aPt:= -1;
  // Init start values
  for i:= 0 to 3 do begin
    nPts[i]:= 2;
    ptX[i, 1]:= 0;
    ptX[i, 2]:= 255;
    ptY[i, 1]:= 0;
    ptY[i, 2]:= 255;
    for j:= 0 to 255 do LUT[i, j]:= j;
  end;
  DrawCurve;
  CurvesPath:= ExtractFileDir(Application.ExeName) + '\Curves\';
  GetCurvesPreset;
end;

procedure TfrmCurves.FormDestroy(Sender: TObject);
begin
  Img.Free;
  TheImage.Free;
end;

procedure TfrmCurves.FormResize(Sender: TObject);
begin
  if ImgLoaded then MakePreviewImage;
end;

end.
